import logging
from typing import Any

from wrapt import wrap_function_wrapper  # type: ignore

from payi.types import IngestUnitsParams
from payi.types.ingest_units_params import Units

from .instrument import PayiInstrumentor


class AnthropicIntrumentor:
    @staticmethod
    def instrument(instrumentor: PayiInstrumentor) -> None:
        try:
            import anthropic  # type: ignore #  noqa: F401  I001

            # wrap_function_wrapper(
            #     "anthropic.resources.completions",
            #     "Completions.create",
            #     chat_wrapper(instrumentor),
            # )

            wrap_function_wrapper(
                "anthropic.resources.messages",
                "Messages.create",
                chat_wrapper(instrumentor),
            )

            wrap_function_wrapper(
                "anthropic.resources.messages",
                "Messages.stream",
                chat_wrapper(instrumentor),
            )

        except Exception as e:
            logging.debug(f"Error instrumenting anthropic: {e}")
            return


@PayiInstrumentor.payi_wrapper
def chat_wrapper(
    instrumentor: PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    args: Any,
    kwargs: Any,
) -> Any:
    return instrumentor.chat_wrapper(
        "system.anthropic",
        process_chunk,
        process_synchronous_response,
        wrapped,
        instance,
        args,
        kwargs,
    )


def process_chunk(chunk: Any, ingest: IngestUnitsParams) -> None:
    if chunk.type == "message_start":
        usage = chunk.message.usage
        units = ingest["units"]

        units["text"] = Units(input=usage.input_tokens, output=0)

        if hasattr(usage, "cache_creation_input_tokens") and usage.cache_creation_input_tokens > 0:
            text_cache_write = usage.cache_creation_input_tokens
            units["text_cache_write"] = Units(input=text_cache_write, output=0)

        if hasattr(usage, "cache_read_input_tokens") and usage.cache_read_input_tokens > 0:
            text_cache_read = usage.cache_read_input_tokens
            units["text_cache_read"] = Units(input=text_cache_read, output=0)

    elif chunk.type == "message_delta":
        usage = chunk.usage
        ingest["units"]["text"]["output"] = usage.output_tokens


def process_synchronous_response(response: Any, ingest: IngestUnitsParams, log_prompt_and_response: bool) -> None:
    usage = response.usage
    input = usage.input_tokens
    ouptut = usage.output_tokens
    units: dict[str, Units] = ingest["units"]

    if hasattr(usage, "cache_creation_input_tokens") and usage.cache_creation_input_tokens > 0:
        text_cache_write = usage.cache_creation_input_tokens
        units["text_cache_write"] = Units(input=text_cache_write, output=0)

    if hasattr(usage, "cache_read_input_tokens") and usage.cache_read_input_tokens > 0:
        text_cache_read = usage.cache_read_input_tokens
        units["text_cache_read"] = Units(input=text_cache_read, output=0)

    units["text"] = Units(input=input, output=ouptut)

    if log_prompt_and_response:
        ingest["provider_response_json"] = response.to_json()



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from payi import Payi, AsyncPayi
from payi.types import (
    CategoryListResponse,
    CategoryDeleteResponse,
    CategoryListResourcesResponse,
    CategoryDeleteResourceResponse,
)
from tests.utils import assert_matches_type

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestCategories:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_list(self, client: Payi) -> None:
        category = client.categories.list()
        assert_matches_type(CategoryListResponse, category, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Payi) -> None:
        response = client.categories.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = response.parse()
        assert_matches_type(CategoryListResponse, category, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Payi) -> None:
        with client.categories.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = response.parse()
            assert_matches_type(CategoryListResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_delete(self, client: Payi) -> None:
        category = client.categories.delete(
            "category",
        )
        assert_matches_type(CategoryDeleteResponse, category, path=["response"])

    @parametrize
    def test_raw_response_delete(self, client: Payi) -> None:
        response = client.categories.with_raw_response.delete(
            "category",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = response.parse()
        assert_matches_type(CategoryDeleteResponse, category, path=["response"])

    @parametrize
    def test_streaming_response_delete(self, client: Payi) -> None:
        with client.categories.with_streaming_response.delete(
            "category",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = response.parse()
            assert_matches_type(CategoryDeleteResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_delete(self, client: Payi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `category` but received ''"):
            client.categories.with_raw_response.delete(
                "",
            )

    @parametrize
    def test_method_delete_resource(self, client: Payi) -> None:
        category = client.categories.delete_resource(
            resource="resource",
            category="category",
        )
        assert_matches_type(CategoryDeleteResourceResponse, category, path=["response"])

    @parametrize
    def test_raw_response_delete_resource(self, client: Payi) -> None:
        response = client.categories.with_raw_response.delete_resource(
            resource="resource",
            category="category",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = response.parse()
        assert_matches_type(CategoryDeleteResourceResponse, category, path=["response"])

    @parametrize
    def test_streaming_response_delete_resource(self, client: Payi) -> None:
        with client.categories.with_streaming_response.delete_resource(
            resource="resource",
            category="category",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = response.parse()
            assert_matches_type(CategoryDeleteResourceResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_delete_resource(self, client: Payi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `category` but received ''"):
            client.categories.with_raw_response.delete_resource(
                resource="resource",
                category="",
            )

        with pytest.raises(ValueError, match=r"Expected a non-empty value for `resource` but received ''"):
            client.categories.with_raw_response.delete_resource(
                resource="",
                category="category",
            )

    @parametrize
    def test_method_list_resources(self, client: Payi) -> None:
        category = client.categories.list_resources(
            "category",
        )
        assert_matches_type(CategoryListResourcesResponse, category, path=["response"])

    @parametrize
    def test_raw_response_list_resources(self, client: Payi) -> None:
        response = client.categories.with_raw_response.list_resources(
            "category",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = response.parse()
        assert_matches_type(CategoryListResourcesResponse, category, path=["response"])

    @parametrize
    def test_streaming_response_list_resources(self, client: Payi) -> None:
        with client.categories.with_streaming_response.list_resources(
            "category",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = response.parse()
            assert_matches_type(CategoryListResourcesResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_list_resources(self, client: Payi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `category` but received ''"):
            client.categories.with_raw_response.list_resources(
                "",
            )


class TestAsyncCategories:
    parametrize = pytest.mark.parametrize("async_client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    async def test_method_list(self, async_client: AsyncPayi) -> None:
        category = await async_client.categories.list()
        assert_matches_type(CategoryListResponse, category, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncPayi) -> None:
        response = await async_client.categories.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = await response.parse()
        assert_matches_type(CategoryListResponse, category, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncPayi) -> None:
        async with async_client.categories.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = await response.parse()
            assert_matches_type(CategoryListResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_delete(self, async_client: AsyncPayi) -> None:
        category = await async_client.categories.delete(
            "category",
        )
        assert_matches_type(CategoryDeleteResponse, category, path=["response"])

    @parametrize
    async def test_raw_response_delete(self, async_client: AsyncPayi) -> None:
        response = await async_client.categories.with_raw_response.delete(
            "category",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = await response.parse()
        assert_matches_type(CategoryDeleteResponse, category, path=["response"])

    @parametrize
    async def test_streaming_response_delete(self, async_client: AsyncPayi) -> None:
        async with async_client.categories.with_streaming_response.delete(
            "category",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = await response.parse()
            assert_matches_type(CategoryDeleteResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_delete(self, async_client: AsyncPayi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `category` but received ''"):
            await async_client.categories.with_raw_response.delete(
                "",
            )

    @parametrize
    async def test_method_delete_resource(self, async_client: AsyncPayi) -> None:
        category = await async_client.categories.delete_resource(
            resource="resource",
            category="category",
        )
        assert_matches_type(CategoryDeleteResourceResponse, category, path=["response"])

    @parametrize
    async def test_raw_response_delete_resource(self, async_client: AsyncPayi) -> None:
        response = await async_client.categories.with_raw_response.delete_resource(
            resource="resource",
            category="category",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = await response.parse()
        assert_matches_type(CategoryDeleteResourceResponse, category, path=["response"])

    @parametrize
    async def test_streaming_response_delete_resource(self, async_client: AsyncPayi) -> None:
        async with async_client.categories.with_streaming_response.delete_resource(
            resource="resource",
            category="category",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = await response.parse()
            assert_matches_type(CategoryDeleteResourceResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_delete_resource(self, async_client: AsyncPayi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `category` but received ''"):
            await async_client.categories.with_raw_response.delete_resource(
                resource="resource",
                category="",
            )

        with pytest.raises(ValueError, match=r"Expected a non-empty value for `resource` but received ''"):
            await async_client.categories.with_raw_response.delete_resource(
                resource="",
                category="category",
            )

    @parametrize
    async def test_method_list_resources(self, async_client: AsyncPayi) -> None:
        category = await async_client.categories.list_resources(
            "category",
        )
        assert_matches_type(CategoryListResourcesResponse, category, path=["response"])

    @parametrize
    async def test_raw_response_list_resources(self, async_client: AsyncPayi) -> None:
        response = await async_client.categories.with_raw_response.list_resources(
            "category",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        category = await response.parse()
        assert_matches_type(CategoryListResourcesResponse, category, path=["response"])

    @parametrize
    async def test_streaming_response_list_resources(self, async_client: AsyncPayi) -> None:
        async with async_client.categories.with_streaming_response.list_resources(
            "category",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            category = await response.parse()
            assert_matches_type(CategoryListResourcesResponse, category, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_list_resources(self, async_client: AsyncPayi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `category` but received ''"):
            await async_client.categories.with_raw_response.list_resources(
                "",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .requests import (
    RequestsResource,
    AsyncRequestsResource,
    RequestsResourceWithRawResponse,
    AsyncRequestsResourceWithRawResponse,
    RequestsResourceWithStreamingResponse,
    AsyncRequestsResourceWithStreamingResponse,
)
from .properties import (
    PropertiesResource,
    AsyncPropertiesResource,
    PropertiesResourceWithRawResponse,
    AsyncPropertiesResourceWithRawResponse,
    PropertiesResourceWithStreamingResponse,
    AsyncPropertiesResourceWithStreamingResponse,
)

__all__ = [
    "PropertiesResource",
    "AsyncPropertiesResource",
    "PropertiesResourceWithRawResponse",
    "AsyncPropertiesResourceWithRawResponse",
    "PropertiesResourceWithStreamingResponse",
    "AsyncPropertiesResourceWithStreamingResponse",
    "RequestsResource",
    "AsyncRequestsResource",
    "RequestsResourceWithRawResponse",
    "AsyncRequestsResourceWithRawResponse",
    "RequestsResourceWithStreamingResponse",
    "AsyncRequestsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from ..._compat import cached_property
from .properties import (
    PropertiesResource,
    AsyncPropertiesResource,
    PropertiesResourceWithRawResponse,
    AsyncPropertiesResourceWithRawResponse,
    PropertiesResourceWithStreamingResponse,
    AsyncPropertiesResourceWithStreamingResponse,
)
from ..._resource import SyncAPIResource, AsyncAPIResource

__all__ = ["RequestsResource", "AsyncRequestsResource"]


class RequestsResource(SyncAPIResource):
    @cached_property
    def properties(self) -> PropertiesResource:
        return PropertiesResource(self._client)

    @cached_property
    def with_raw_response(self) -> RequestsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return RequestsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> RequestsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return RequestsResourceWithStreamingResponse(self)


class AsyncRequestsResource(AsyncAPIResource):
    @cached_property
    def properties(self) -> AsyncPropertiesResource:
        return AsyncPropertiesResource(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncRequestsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return AsyncRequestsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncRequestsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return AsyncRequestsResourceWithStreamingResponse(self)


class RequestsResourceWithRawResponse:
    def __init__(self, requests: RequestsResource) -> None:
        self._requests = requests

    @cached_property
    def properties(self) -> PropertiesResourceWithRawResponse:
        return PropertiesResourceWithRawResponse(self._requests.properties)


class AsyncRequestsResourceWithRawResponse:
    def __init__(self, requests: AsyncRequestsResource) -> None:
        self._requests = requests

    @cached_property
    def properties(self) -> AsyncPropertiesResourceWithRawResponse:
        return AsyncPropertiesResourceWithRawResponse(self._requests.properties)


class RequestsResourceWithStreamingResponse:
    def __init__(self, requests: RequestsResource) -> None:
        self._requests = requests

    @cached_property
    def properties(self) -> PropertiesResourceWithStreamingResponse:
        return PropertiesResourceWithStreamingResponse(self._requests.properties)


class AsyncRequestsResourceWithStreamingResponse:
    def __init__(self, requests: AsyncRequestsResource) -> None:
        self._requests = requests

    @cached_property
    def properties(self) -> AsyncPropertiesResourceWithStreamingResponse:
        return AsyncPropertiesResourceWithStreamingResponse(self._requests.properties)



import json
import logging
from typing import Any
from importlib.metadata import version

from wrapt import wrap_function_wrapper  # type: ignore

from payi.types import IngestUnitsParams
from payi.types.ingest_units_params import Units

from .instrument import PayiInstrumentor


class OpenAiInstrumentor:
    @staticmethod
    def instrument(instrumentor: PayiInstrumentor) -> None:
        try:
            from openai import OpenAI  # type: ignore #  noqa: F401  I001

            wrap_function_wrapper(
                "openai.resources.chat.completions",
                "Completions.create",
                chat_wrapper(instrumentor),
            )
        except Exception as e:
            logging.debug(f"Error instrumenting openai: {e}")
            return


@PayiInstrumentor.payi_wrapper
def chat_wrapper(
    instrumentor: PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    args: Any,
    kwargs: Any,
) -> Any:
    return instrumentor.chat_wrapper(
        "system.openai",
        process_chat_chunk,
        process_chat_synchronous_response,
        wrapped,
        instance,
        args,
        kwargs,
    )


def process_chat_synchronous_response(response: str, ingest: IngestUnitsParams, log_prompt_and_response: bool) -> None:
    response_dict = model_to_dict(response)

    add_usage_units(response_dict["usage"], ingest["units"])

    if log_prompt_and_response:
        ingest["provider_response_json"] = [json.dumps(response_dict)]


def process_chat_chunk(chunk: Any, ingest: IngestUnitsParams) -> None:
    model = model_to_dict(chunk)
    usage = model.get("usage")
    if usage:
        add_usage_units(usage, ingest["units"])


def model_to_dict(model: Any) -> Any:
    if version("pydantic") < "2.0.0":
        return model.dict()
    if hasattr(model, "model_dump"):
        return model.model_dump()
    elif hasattr(model, "parse"):  # Raw API response
        return model_to_dict(model.parse())
    else:
        return model


def add_usage_units(usage: "dict[str, Any]", units: "dict[str, Units]") -> None:
    input = usage["prompt_tokens"] if "prompt_tokens" in usage else 0
    output = usage["completion_tokens"] if "completion_tokens" in usage else 0
    input_cache = 0

    prompt_tokens_details = usage.get("prompt_tokens_details")
    if prompt_tokens_details:
        input_cache = prompt_tokens_details.get("cached_tokens", 0)
        if input_cache != 0:
            units["text_cache_read"] = Units(input=input_cache, output=0)

    input -= input_cache

    units["text"] = Units(input=input, output=output)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, Union, Mapping
from typing_extensions import Self, override

import httpx

from . import _exceptions
from ._qs import Querystring
from ._types import (
    NOT_GIVEN,
    Omit,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
)
from ._utils import (
    is_given,
    get_async_library,
)
from ._version import __version__
from .resources import ingest, billing_models, price_modifiers
from ._streaming import Stream as Stream, AsyncStream as AsyncStream
from ._exceptions import PayiError, APIStatusError
from ._base_client import (
    DEFAULT_MAX_RETRIES,
    SyncAPIClient,
    AsyncAPIClient,
)
from .resources.limits import limits
from .resources.requests import requests
from .resources.categories import categories
from .resources.experiences import experiences

__all__ = ["Timeout", "Transport", "ProxiesTypes", "RequestOptions", "Payi", "AsyncPayi", "Client", "AsyncClient"]


class Payi(SyncAPIClient):
    limits: limits.LimitsResource
    ingest: ingest.IngestResource
    categories: categories.CategoriesResource
    experiences: experiences.ExperiencesResource
    billing_models: billing_models.BillingModelsResource
    price_modifiers: price_modifiers.PriceModifiersResource
    requests: requests.RequestsResource
    with_raw_response: PayiWithRawResponse
    with_streaming_response: PayiWithStreamedResponse

    # client options
    api_key: str

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous payi client instance.

        This automatically infers the `api_key` argument from the `PAYI_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("PAYI_API_KEY")
        if api_key is None:
            raise PayiError(
                "The api_key client option must be set either by passing api_key to the client or by setting the PAYI_API_KEY environment variable"
            )
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("PAYI_BASE_URL")
        if base_url is None:
            base_url = f"https://api.pay-i.com"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self.limits = limits.LimitsResource(self)
        self.ingest = ingest.IngestResource(self)
        self.categories = categories.CategoriesResource(self)
        self.experiences = experiences.ExperiencesResource(self)
        self.billing_models = billing_models.BillingModelsResource(self)
        self.price_modifiers = price_modifiers.PriceModifiersResource(self)
        self.requests = requests.RequestsResource(self)
        self.with_raw_response = PayiWithRawResponse(self)
        self.with_streaming_response = PayiWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"xProxy-api-key": api_key}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class AsyncPayi(AsyncAPIClient):
    limits: limits.AsyncLimitsResource
    ingest: ingest.AsyncIngestResource
    categories: categories.AsyncCategoriesResource
    experiences: experiences.AsyncExperiencesResource
    billing_models: billing_models.AsyncBillingModelsResource
    price_modifiers: price_modifiers.AsyncPriceModifiersResource
    requests: requests.AsyncRequestsResource
    with_raw_response: AsyncPayiWithRawResponse
    with_streaming_response: AsyncPayiWithStreamedResponse

    # client options
    api_key: str

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultAsyncHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async payi client instance.

        This automatically infers the `api_key` argument from the `PAYI_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("PAYI_API_KEY")
        if api_key is None:
            raise PayiError(
                "The api_key client option must be set either by passing api_key to the client or by setting the PAYI_API_KEY environment variable"
            )
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("PAYI_BASE_URL")
        if base_url is None:
            base_url = f"https://api.pay-i.com"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self.limits = limits.AsyncLimitsResource(self)
        self.ingest = ingest.AsyncIngestResource(self)
        self.categories = categories.AsyncCategoriesResource(self)
        self.experiences = experiences.AsyncExperiencesResource(self)
        self.billing_models = billing_models.AsyncBillingModelsResource(self)
        self.price_modifiers = price_modifiers.AsyncPriceModifiersResource(self)
        self.requests = requests.AsyncRequestsResource(self)
        self.with_raw_response = AsyncPayiWithRawResponse(self)
        self.with_streaming_response = AsyncPayiWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"xProxy-api-key": api_key}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": f"async:{get_async_library()}",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class PayiWithRawResponse:
    def __init__(self, client: Payi) -> None:
        self.limits = limits.LimitsResourceWithRawResponse(client.limits)
        self.ingest = ingest.IngestResourceWithRawResponse(client.ingest)
        self.categories = categories.CategoriesResourceWithRawResponse(client.categories)
        self.experiences = experiences.ExperiencesResourceWithRawResponse(client.experiences)
        self.billing_models = billing_models.BillingModelsResourceWithRawResponse(client.billing_models)
        self.price_modifiers = price_modifiers.PriceModifiersResourceWithRawResponse(client.price_modifiers)
        self.requests = requests.RequestsResourceWithRawResponse(client.requests)


class AsyncPayiWithRawResponse:
    def __init__(self, client: AsyncPayi) -> None:
        self.limits = limits.AsyncLimitsResourceWithRawResponse(client.limits)
        self.ingest = ingest.AsyncIngestResourceWithRawResponse(client.ingest)
        self.categories = categories.AsyncCategoriesResourceWithRawResponse(client.categories)
        self.experiences = experiences.AsyncExperiencesResourceWithRawResponse(client.experiences)
        self.billing_models = billing_models.AsyncBillingModelsResourceWithRawResponse(client.billing_models)
        self.price_modifiers = price_modifiers.AsyncPriceModifiersResourceWithRawResponse(client.price_modifiers)
        self.requests = requests.AsyncRequestsResourceWithRawResponse(client.requests)


class PayiWithStreamedResponse:
    def __init__(self, client: Payi) -> None:
        self.limits = limits.LimitsResourceWithStreamingResponse(client.limits)
        self.ingest = ingest.IngestResourceWithStreamingResponse(client.ingest)
        self.categories = categories.CategoriesResourceWithStreamingResponse(client.categories)
        self.experiences = experiences.ExperiencesResourceWithStreamingResponse(client.experiences)
        self.billing_models = billing_models.BillingModelsResourceWithStreamingResponse(client.billing_models)
        self.price_modifiers = price_modifiers.PriceModifiersResourceWithStreamingResponse(client.price_modifiers)
        self.requests = requests.RequestsResourceWithStreamingResponse(client.requests)


class AsyncPayiWithStreamedResponse:
    def __init__(self, client: AsyncPayi) -> None:
        self.limits = limits.AsyncLimitsResourceWithStreamingResponse(client.limits)
        self.ingest = ingest.AsyncIngestResourceWithStreamingResponse(client.ingest)
        self.categories = categories.AsyncCategoriesResourceWithStreamingResponse(client.categories)
        self.experiences = experiences.AsyncExperiencesResourceWithStreamingResponse(client.experiences)
        self.billing_models = billing_models.AsyncBillingModelsResourceWithStreamingResponse(client.billing_models)
        self.price_modifiers = price_modifiers.AsyncPriceModifiersResourceWithStreamingResponse(client.price_modifiers)
        self.requests = requests.AsyncRequestsResourceWithStreamingResponse(client.requests)


Client = Payi

AsyncClient = AsyncPayi



import json
import uuid
import asyncio
import inspect
import logging
import traceback
from typing import Any, Set, Union, Callable, Optional

from wrapt import ObjectProxy  # type: ignore

from payi import Payi, AsyncPayi
from payi.types import IngestUnitsParams

from .Stopwatch import Stopwatch
from .Instruments import Instruments


class PayiInstrumentor:
    def __init__(
        self,
        payi: Union[Payi, AsyncPayi, None] = None,
        instruments: Union[Set[Instruments], None] = None,
        log_prompt_and_response: bool = True,
        prompt_and_response_logger: Optional[
            Callable[[str, "dict[str, str]"], None]
        ] = None,  # (request id, dict of data to store) -> None
    ):
        self._payi: Union[Payi, AsyncPayi, None] = payi
        self._context_stack: list[dict[str, Any]] = []  # Stack of context dictionaries
        self._log_prompt_and_response: bool = log_prompt_and_response
        self._prompt_and_response_logger: Optional[Callable[[str, dict[str, str]], None]] = prompt_and_response_logger

        self._blocked_limits: set[str] = set()
        self._exceeded_limits: set[str] = set()

        if instruments is None or Instruments.ALL in instruments:
            self._instrument_all()
        else:
            self._instrument_specific(instruments)

    def _instrument_all(self) -> None:
        self._instrument_openai()
        self._instrument_anthropic()

    def _instrument_specific(self, instruments: Set[Instruments]) -> None:
        if Instruments.OPENAI in instruments:
            self._instrument_openai()
        if Instruments.ANTHROPIC in instruments:
            self._instrument_anthropic()

    def _instrument_openai(self) -> None:
        from .OpenAIInstrumentor import OpenAiInstrumentor

        try:
            OpenAiInstrumentor.instrument(self)

        except Exception as e:
            logging.error(f"Error instrumenting OpenAI: {e}")

    def _instrument_anthropic(self) -> None:
        from .AnthropicInstrumentor import AnthropicIntrumentor

        try:
            AnthropicIntrumentor.instrument(self)

        except Exception as e:
            logging.error(f"Error instrumenting Anthropic: {e}")

    def _ingest_units(self, ingest_units: IngestUnitsParams) -> None:
        # return early if there are no units to ingest and on a successul ingest request
        if int(ingest_units.get("http_status_code") or 0) < 400:
            units = ingest_units.get("units", {})
            if not units or all(unit.get("input", 0) == 0 and unit.get("output", 0) == 0 for unit in units.values()):
                logging.error(
                    'No units to ingest.  For OpenAI streaming calls, make sure you pass stream_options={"include_usage": True}'
                )
                return

        try:
            if isinstance(self._payi, AsyncPayi):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                try:
                    ingest_result = loop.run_until_complete(self._payi.ingest.units(**ingest_units))
                finally:
                    loop.close()
            elif isinstance(self._payi, Payi):
                ingest_result = self._payi.ingest.units(**ingest_units)
            else:
                logging.error("No payi instance to ingest units")
                return

            if ingest_result.xproxy_result.limits:
                for limit_id, state in ingest_result.xproxy_result.limits.items():
                    removeBlockedId: bool = False

                    if state.state == "blocked":
                        self._blocked_limits.add(limit_id)
                    elif state.state == "exceeded":
                        self._exceeded_limits.add(limit_id)
                        removeBlockedId = True
                    elif state.state == "ok":
                        removeBlockedId = True

                    # opportunistically remove blocked limits
                    if removeBlockedId:
                        self._blocked_limits.discard(limit_id)

            if self._log_prompt_and_response and self._prompt_and_response_logger:
                request_id = ingest_result.xproxy_result.request_id

                log_data = {}
                response_json = ingest_units.pop("provider_response_json", None)
                request_json = ingest_units.pop("provider_request_json", None)
                stack_trace = ingest_units.get("properties", {}).pop("system.stack_trace", None)  # type: ignore

                if response_json is not None:
                    # response_json is a list of strings, convert a single json string
                    log_data["provider_response_json"] = json.dumps(response_json)
                if request_json is not None:
                    log_data["provider_request_json"] = request_json
                if stack_trace is not None:
                    log_data["stack_trace"] = stack_trace

                self._prompt_and_response_logger(request_id, log_data)  # type: ignore

        except Exception as e:
            logging.error(f"Error Pay-i ingesting result: {e}")

    def _call_func(
        self,
        func: Any,
        proxy: bool,
        limit_ids: Optional["list[str]"],
        request_tags: Optional["list[str]"],
        experience_name: Optional[str],
        experience_id: Optional[str],
        user_id: Optional[str],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        if len(self._context_stack) > 0:
            # copy current context into the upcoming context
            context = self._context_stack[-1].copy()
            context.pop("proxy", None)
            previous_experience_name = context["experience_name"]
            previous_experience_id = context["experience_id"]
        else:
            context = {}
            previous_experience_name = None
            previous_experience_id = None

        with self:
            context["proxy"] = proxy

            # Handle experience name and ID logic
            if not experience_name:
                # If no experience_name specified, use previous values
                context["experience_name"] = previous_experience_name
                context["experience_id"] = previous_experience_id
            else:
                # If experience_name is specified
                if experience_name == previous_experience_name:
                    # Same experience name, use previous ID unless new one specified
                    context["experience_name"] = experience_name
                    context["experience_id"] = experience_id if experience_id else previous_experience_id
                else:
                    # Different experience name, use specified ID or generate one
                    context["experience_name"] = experience_name
                    context["experience_id"] = experience_id if experience_id else str(uuid.uuid4())

            # set any values explicitly passed by the caller, otherwise use what is already in the context
            if limit_ids:
                context["limit_ids"] = limit_ids
            if request_tags:
                context["request_tags"] = request_tags
            if user_id:
                context["user_id"] = user_id

            self.set_context(context)

            return func(*args, **kwargs)

    def __enter__(self) -> Any:
        # Push a new context dictionary onto the stack
        self._context_stack.append({})
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        # Pop the current context off the stack
        if self._context_stack:
            self._context_stack.pop()

    def set_context(self, context: "dict[str, Any]") -> None:
        # Update the current top of the stack with the provided context
        if self._context_stack:
            self._context_stack[-1].update(context)

    def get_context(self) -> Optional["dict[str, Any]"]:
        # Return the current top of the stack
        return self._context_stack[-1] if self._context_stack else None

    def chat_wrapper(
        self,
        category: str,
        process_chunk: Callable[[Any, IngestUnitsParams], None],
        process_synchronous_response: Optional[Callable[[Any, IngestUnitsParams, bool], None]],
        wrapped: Any,
        instance: Any,
        args: Any,
        kwargs: Any,
    ) -> Any:
        context = self.get_context()

        if not context:
            # should not happen
            return wrapped(*args, **kwargs)

        if context.get("proxy", True):
            proxy_extra_headers = kwargs.get("extra_headers", {})

            self._update_headers(context, proxy_extra_headers)

            if "extra_headers" not in kwargs:
                kwargs["extra_headers"] = proxy_extra_headers

            return wrapped(*args, **kwargs)

        ingest: IngestUnitsParams = {"category": category, "resource": kwargs.get("model"), "units": {}}

        # blocked_limit = next((limit for limit in (context.get('limit_ids') or []) if limit in self._blocked_limits), None)
        # if blocked_limit:
        #      raise Exception(f"Limit {blocked_limit} is blocked")
        current_frame = inspect.currentframe()
        # f_back excludes the current frame, strip() cleans up whitespace and newlines
        stack = [frame.strip() for frame in traceback.format_stack(current_frame.f_back)]  # type: ignore

        # TODO add back once feature is in prod
        # ingest['properties'] = { 'system.stack_trace': json.dumps(stack) }

        sw = Stopwatch()
        stream = kwargs.get("stream", False)

        try:
            limit_ids = context.get("limit_ids")
            request_tags = context.get("request_tags")
            experience_name = context.get("experience_name")
            experience_id = context.get("experience_id")
            user_id = context.get("user_id")

            if limit_ids:
                ingest["limit_ids"] = limit_ids
            if request_tags:
                ingest["request_tags"] = request_tags
            if experience_name:
                ingest["experience_name"] = experience_name
            if experience_id:
                ingest["experience_id"] = experience_id
            if user_id:
                ingest["user_id"] = user_id

            extra_headers: dict[str, str] = kwargs.get("extra_headers") or {}

            if len(extra_headers) > 0:
                ingest["provider_request_headers"] = {k: [v] for k, v in extra_headers.items()}  # type: ignore

            provider_prompt = {}
            for k, v in kwargs.items():
                if k == "messages":
                    provider_prompt[k] = [m.model_dump() if hasattr(m, "model_dump") else m for m in v]
                elif k in ["extra_headers", "extra_query"]:
                    pass
                else:
                    provider_prompt[k] = v

            if self._log_prompt_and_response:
                ingest["provider_request_json"] = json.dumps(provider_prompt)

            sw.start()
            response = wrapped(*args, **kwargs.copy())

        except Exception as e:  # pylint: disable=broad-except
            sw.stop()
            duration = sw.elapsed_ms_int()

            # TODO ingest error

            raise e

        if stream:
            return ChatStreamWrapper(
                response=response,
                instance=instance,
                instrumentor=self,
                log_prompt_and_response=self._log_prompt_and_response,
                ingest=ingest,
                stopwatch=sw,
                process_chunk=process_chunk,
            )

        sw.stop()
        duration = sw.elapsed_ms_int()
        ingest["end_to_end_latency_ms"] = duration
        ingest["http_status_code"] = 200

        if process_synchronous_response:
            process_synchronous_response(response, ingest, self._log_prompt_and_response)

        self._ingest_units(ingest)

        return response

    @staticmethod
    def _update_headers(
        context: "dict[str, Any]",
        extra_headers: "dict[str, str]",
    ) -> None:
        limit_ids: Optional[list[str]] = context.get("limit_ids")
        request_tags: Optional[list[str]] = context.get("request_tags")
        experience_name: Optional[str] = context.get("experience_name")
        experience_id: Optional[str] = context.get("experience_id")
        user_id: Optional[str] = context.get("user_id")

        if limit_ids is not None:
            existing_limit_ids = extra_headers.get("xProxy-Limit-IDs")
            limit_ids_str = ",".join(limit_ids)
            if existing_limit_ids is None:
                extra_headers["xProxy-Limit-IDs"] = limit_ids_str
            else:
                extra_headers["xProxy-Limit-IDs"] = f"{existing_limit_ids},{limit_ids_str}"

        if request_tags is not None:
            existing_request_tags = extra_headers.get("xProxy-Request-Tags")
            request_tags_str = ",".join(request_tags)
            if existing_request_tags is None:
                extra_headers["xProxy-Request-Tags"] = request_tags_str
            else:
                extra_headers["xProxy-Request-Tags"] = f"{existing_request_tags},{request_tags_str}"

        if experience_name is not None:
            extra_headers["xProxy-Experience-Name"] = experience_name

        if experience_id is not None:
            extra_headers["xProxy-Experience-ID"] = experience_id

        if user_id is not None:
            extra_headers["xProxy-User-ID"] = user_id

    @staticmethod
    def payi_wrapper(func: Any) -> Any:
        def _payi_wrapper(o: Any) -> Any:
            def wrapper(wrapped: Any, instance: Any, args: Any, kwargs: Any) -> Any:
                return func(
                    o,
                    wrapped,
                    instance,
                    args,
                    kwargs,
                )

            return wrapper

        return _payi_wrapper


class ChatStreamWrapper(ObjectProxy):  # type: ignore
    def __init__(
        self,
        response: Any,
        instance: Any,
        instrumentor: PayiInstrumentor,
        ingest: IngestUnitsParams,
        stopwatch: Stopwatch,
        process_chunk: Optional[Callable[[Any, IngestUnitsParams], None]] = None,
        log_prompt_and_response: bool = True,
    ) -> None:
        super().__init__(response)  # type: ignore

        self._response = response
        self._instance = instance

        self._instrumentor = instrumentor
        self._stopwatch: Stopwatch = stopwatch
        self._ingest: IngestUnitsParams = ingest
        self._log_prompt_and_response: bool = log_prompt_and_response
        self._responses: list[str] = []

        self._process_chunk: Optional[Callable[[Any, IngestUnitsParams], None]] = process_chunk

        self._first_token: bool = True

    def __enter__(self) -> Any:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None: 
        self.__wrapped__.__exit__(exc_type, exc_val, exc_tb)  # type: ignore

    async def __aenter__(self) -> Any:
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.__wrapped__.__aexit__(exc_type, exc_val, exc_tb)  # type: ignore

    def __iter__(self) -> Any:
        return self

    def __aiter__(self) -> Any:
        return self

    def __next__(self) -> Any:
        try:
            chunk: Any = self.__wrapped__.__next__()  # type: ignore
        except Exception as e:
            if isinstance(e, StopIteration):
                self._stop_iteration()
            raise e
        else:
            self._evaluate_chunk(chunk)
            return chunk

    async def __anext__(self) -> Any:
        try:
            chunk: Any = await self.__wrapped__.__anext__()  # type: ignore
        except Exception as e:
            if isinstance(e, StopAsyncIteration):
                self._stop_iteration()
            raise e
        else:
            self._evaluate_chunk(chunk)
            return chunk

    def _evaluate_chunk(self, chunk: Any) -> None:
        if self._first_token:
            self._ingest["time_to_first_token_ms"] = self._stopwatch.elapsed_ms_int()
            self._first_token = False

        if self._log_prompt_and_response:
            self._responses.append(chunk.to_json())

        if self._process_chunk:
            self._process_chunk(chunk, self._ingest)

    def _stop_iteration(self) -> None:
        self._stopwatch.stop()
        self._ingest["end_to_end_latency_ms"] = self._stopwatch.elapsed_ms_int()
        self._ingest["http_status_code"] = 200

        if self._log_prompt_and_response:
            self._ingest["provider_response_json"] = self._responses

        self._instrumentor._ingest_units(self._ingest)


global _instrumentor
_instrumentor: PayiInstrumentor


def payi_instrument(
    payi: Optional[Union[Payi, AsyncPayi]] = None,
    instruments: Optional[Set[Instruments]] = None,
    log_prompt_and_response: bool = True,
    prompt_and_response_logger: Optional[Callable[[str, "dict[str, str]"], None]] = None,
) -> None:
    global _instrumentor
    _instrumentor = PayiInstrumentor(
        payi=payi,
        instruments=instruments,
        log_prompt_and_response=log_prompt_and_response,
        prompt_and_response_logger=prompt_and_response_logger,
    )


def ingest(
    limit_ids: Optional["list[str]"] = None,
    request_tags: Optional["list[str]"] = None,
    experience_name: Optional[str] = None,
    experience_id: Optional[str] = None,
    user_id: Optional[str] = None,
) -> Any:
    def _ingest(func: Any) -> Any:
        def _ingest_wrapper(*args: Any, **kwargs: Any) -> Any:
            return _instrumentor._call_func(
                func,
                False,  # false -> ingest
                limit_ids,
                request_tags,
                experience_name,
                experience_id,
                user_id,
                *args,
                **kwargs,
            )

        return _ingest_wrapper

    return _ingest


def proxy(
    limit_ids: Optional["list[str]"] = None,
    request_tags: Optional["list[str]"] = None,
    experience_name: Optional[str] = None,
    experience_id: Optional[str] = None,
    user_id: Optional[str] = None,
) -> Any:
    def _proxy(func: Any) -> Any:
        def _proxy_wrapper(*args: Any, **kwargs: Any) -> Any:
            return _instrumentor._call_func(
                func, True, limit_ids, request_tags, experience_name, experience_id, user_id, *args, **kwargs
            )

        return _proxy_wrapper

    return _proxy



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union, Iterable, Optional
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["IngestEventParam", "Units", "ProviderRequestHeader", "ProviderResponseHeader"]


class Units(TypedDict, total=False):
    input: int

    output: int


class ProviderRequestHeader(TypedDict, total=False):
    name: Required[str]

    value: Optional[str]


class ProviderResponseHeader(TypedDict, total=False):
    name: Required[str]

    value: Optional[str]


class IngestEventParam(TypedDict, total=False):
    category: Required[str]

    resource: Required[str]

    units: Required[Dict[str, Units]]

    csat_rating: Optional[int]

    end_to_end_latency_ms: Optional[int]

    event_timestamp: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]

    experience_id: Optional[str]

    experience_name: Optional[str]

    experience_properties: Optional[Dict[str, str]]

    http_status_code: Optional[int]

    limit_ids: Optional[List[str]]

    properties: Optional[Dict[str, str]]

    provider_request_headers: Optional[Iterable[ProviderRequestHeader]]

    provider_request_json: Optional[str]

    provider_response_headers: Optional[Iterable[ProviderResponseHeader]]

    provider_response_json: Union[str, List[str], None]

    provider_uri: Optional[str]

    request_tags: Optional[List[str]]

    time_to_first_token_ms: Optional[int]

    user_id: Optional[str]
